/* $Id: nettype.c,v 1.7 2001/09/19 06:15:22 m_sato Exp $ */
/* 
 * nettype.c -- chat with specified node
 *
 * this is test program for "linux packet interface" 
 * 
 * see packet(7) and netdevice(7)
 */
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h> 
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/icmp.h>
#include <linux/version.h>
#include <net/if.h>
#include <net/if_arp.h>
/*
#include <net/ip.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/icmp.h>
*/
#include "rawsock_utils.h"

#define IP_OFFSET    0x1FFF          /* "Fragment Offset" part       */
#define NIPQUAD(addr) \
        ((unsigned char *)&addr)[0], \
        ((unsigned char *)&addr)[1], \
        ((unsigned char *)&addr)[2], \
        ((unsigned char *)&addr)[3]


#define MAXPACKETLEN 0x10000	/* is it large enough ? */

#define DEBUG printf

void usage();
void getmsec(int *hh, int *mm, int *sec, int *msec);
int showmessage(struct sockaddr *myaddr, struct sockaddr *peeraddr, unsigned char *buf, int count);
void print_ethaddr(const u_char *p);
void hexdump(unsigned char *buf, int nbytes);

#define print_ip(a)      printf("%u.%u.%u.%u", NIPQUAD(a));

int ifindex;
int showaddress = 0;
int showethaddr = 0;
int showtimestamp = 0;
int bothdisplay = 0;
int debug = 0;
int portno = 25; /* SMTP */

int 
main(int argc, char **argv)
{
	int c;
	int pd;
	char *interface;
	unsigned char buf[2048];	/* must be larger than every MTU */
	unsigned char *p;
	struct iphdr *ip;
	struct tcphdr *tcp;
	struct udphdr *udp;
	struct sockaddr_ll sll;
	int i, offset;
	struct ifreq ifr;
	struct sockaddr myaddr;
	int addrlen;
	int broadcast;
	extern char *optarg;
	char *node;
	char *pattern;
	int send_count, send_size;
	int infinite;
	fd_set rfds, wfds;
	int fd;
	int hh, mm, sec, msec;
	char *filename;

        __u32 src, dst;
        __u16 src_port=0xFFFF, dst_port=0xFFFF;

	interface = NULL;
	broadcast = 0;
	node = NULL;
	pattern = NULL;
	send_count = 1;
	send_size = 0;
	infinite = 1;
	filename = "packetlog";

	for (;;) {
		c = getopt(argc, argv, "aec:f:i:bdtp:");
		switch (c) {
		case 'D':
			debug = 1;
			break;
		case 'i':
			interface = strdup(optarg);
			break;
		case 'c':
			send_count = atoi(optarg);
			break;
		case 'b':
			bothdisplay = 1;
			break;
		case 'e':
			showethaddr = 1;
			break;
		case 'a':
			showaddress = 1;
			break;
		case 't':
			showtimestamp = 1;
			break;
		case 'f':
			filename = strdup(optarg);
			break;
                case -1:  
                        if (argv[optind] && *argv[optind] != '-') {
                                node = strdup(argv[optind]);
                                optind++;
                        } 
			/* FALL THROUGH */
                default:
			if (argv[optind] == NULL)
				goto next;
                        usage();
                        exit(1);
		}
	}
next:
	if (interface == NULL) {
		printf("xxx: no interface or node specified \n");
		usage();
		exit(1);
	}

	pd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
	if (pd == -1) {
		perror("socket():");
		exit(1);
	}
	DEBUG("interface index: ");

	/* get interface index number */
	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, interface, IFNAMSIZ);
	if (ioctl(pd, SIOCGIFINDEX, &ifr) == -1) {
		perror("SIOCGIFINDEX");
		exit(1);
	}
	ifindex = ifr.ifr_ifindex;
	DEBUG("%d\n", ifindex);

	/* get My hardware address */
	DEBUG("My hardware address: ");
	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, interface, IFNAMSIZ);
	if (ioctl(pd, SIOCGIFHWADDR, &ifr) == -1) {
		perror("SIOCGIFINDEX");
		exit(1);
	}
	myaddr = ifr.ifr_hwaddr;

	switch (myaddr.sa_family) {
	case ARPHRD_ETHER:
		addrlen = 6;
		break;
	default:
		addrlen = sizeof(myaddr.sa_data);
	}

	p = myaddr.sa_data;
	for (i = 0; i < addrlen - 1; i++)
		DEBUG("%02x:", *p++);
	DEBUG("%02x\n", *p++);

	DEBUG("binding to the interface ifindex=%d\n", ifindex);
	/*
	 * bind(2) uses only sll_protocol and sll_ifindex. see packet(7)
	 */
	memset(&sll, 0xff, sizeof(sll));
	sll.sll_family = AF_PACKET;	/* allways AF_PACKET */
	sll.sll_protocol = htons(ETH_P_ALL);
	sll.sll_ifindex = ifindex;
	if (bind(pd, (struct sockaddr *)&sll, sizeof sll) == -1) {
		perror("bind():");
		exit(1);
	}

	DEBUG("flush receive buffer\n");

	fd = open(filename, O_CREAT | O_WRONLY, 0666);
	if (fd < 0) {
		perror("open");
	}

	/* 
	 * first, flush all received packets. 
	 *
         * raw-socket receives packets from all interfaces,
	 * while the socket is not binded to a interface
         */
	do {
		fd_set fds;
		struct timeval t;
		FD_ZERO(&fds);	
		FD_SET(pd, &fds);
		memset(&t, 0, sizeof(t));
		i = select(FD_SETSIZE, &fds, NULL, NULL, &t);
		if (i > 0)
			recv(pd, buf, i, 0);
	} while (i);

	while (infinite || send_count) {
		FD_ZERO(&rfds);	
		FD_ZERO(&wfds);	
		FD_SET(pd, &rfds);
		FD_SET(STDIN_FILENO, &rfds);
		if (pattern)
			FD_SET(pd, &wfds);
		i = select(FD_SETSIZE, &rfds, &wfds, NULL, NULL);
		if (i == 0)
			continue;
		if (i == -1) {
			perror("select");
			exit(1);
		}
		if (FD_ISSET(pd, &rfds)) {
			i = recv(pd, buf, sizeof(buf), 0);
			if (i < 0) {
				perror("recv():");
				exit(1);
			}
			if (i == 0)
				continue;
                        else if (!bothdisplay && memcmp(myaddr.sa_data, buf, 6) != 0)
				continue;

			if (showaddress && showtimestamp) {
				getmsec(&hh, &mm, &sec, &msec);
				printf("@%02d:%02d:%02d.%03d ", hh, mm, sec, msec);
			}
			if (showaddress && showethaddr) {
				print_ethaddr(buf);
			}
                        ip = (struct iphdr *)(buf + sizeof(struct ethhdr));
			tcp = (struct tcphdr *)(ip + ip->ihl);
			udp = (struct udphdr *)(ip + ip->ihl);

			src = ip->saddr;
			dst = ip->daddr;
			src_port=0xFFFF;
			dst_port=0xFFFF;

			offset = ntohs(ip->frag_off) & IP_OFFSET;

			if (!offset) {
				if (ip->protocol==IPPROTO_TCP) {
					src_port=ntohs(tcp->source);
					dst_port=ntohs(tcp->dest);
				}
				else if (ip->protocol==IPPROTO_UDP) {
					src_port=ntohs(udp->source);
					dst_port=ntohs(udp->dest);
				}
			}

			if (showaddress) {
				print_ip(ip->saddr);
				printf(":%d -> ", src_port);
				print_ip(ip->daddr);
				printf(":%d\n", dst_port);
			}

			write(fd, buf, sizeof(buf));
		}
	}
	close(fd);
	exit(0);
}

void usage()
{
	fprintf(stderr, "usage: rcv [options] -i interface [-p port# ]\n");
	fprintf(stderr, "            options: -D  debug mode\n");
	fprintf(stderr, "                     -d  display address\n");
	fprintf(stderr, "                     -a  display address\n");
	fprintf(stderr, "                     -t  display timestamp\n");
	fprintf(stderr, "                     -b  Both (Bi-directional) packet\n");
}

void
print_ethaddr(p)
	const u_char *p;
{
	int i;
	struct ethhdr *eh;

	eh = (struct ethhdr *)p;

	for (i = 0; i < 5; ++i) 
		printf("%02x:", (int)eh->h_source[i]);
	printf("%02x -> ", (int)eh->h_source[i]);

	for (i = 0; i < 5; ++i) 
		printf("%02x:", (int)eh->h_dest[i]);
	printf("%02x", (int)eh->h_dest[i]);
	printf("\n");
}

void getmsec(int *hh, int *mm, int *sec, int *msec)
{
  struct timezone tz;
  struct timeval t;

  tz.tz_minuteswest = 0;
  tz.tz_dsttime = 0;

  gettimeofday(&t, &tz);

  *sec = t.tv_sec % 60;
  *hh  = t.tv_sec / 60;
  *mm  = *hh % 60;
  *hh  /= 60;
  *hh  = (*hh + 9) % 24;
  *msec = (t.tv_usec + 500) / 1000;
} 
